#include "bondvalencedialog.h"
#include "ui_bondvalencedialog.h"
#if defined(JAV_WIDGET)
#include "crystalviewwidget.h"
#else
#include "javwindow.h"
#endif
#include "floatdelegate.h"
#include "savedialog.h"
#include "javkeysettings.h"
#include "mainwindow.h"
#include <QMessageBox>

extern "C" void jav_bond_valence_sum_(int *km, int *n, float vet[], int cn[],
                                      float *gii, int *cod);
extern "C" int  jav_bvs_ndinfo_(int *wm, int *kat);
extern "C" void jav_bvs_dinfo_(int *wm, int *kat, int *nd, bond_info_t *dinfo);
extern "C" void jav_export_bv_ (const char *filenam, size_t sz);

BondValenceDialog::BondValenceDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::BondValenceDialog)
{
#if defined(JAV_WIDGET)
    jav = qobject_cast<CrystalViewWidget*>(parent);
#else
    jav = qobject_cast<JavWindow*>(parent);
#endif
    ui->setupUi(this);    

    bvsModel = new QStandardItemModel(this);
    QStringList bvsLabels = {"No.","Label","Type","BVS","CN"};
    bvsModel->setHorizontalHeaderLabels(bvsLabels);
    ui->bvsTable->verticalHeader()->setDefaultSectionSize(22);
    ui->bvsTable->setModel(bvsModel);
    ui->bvsTable->verticalHeader()->hide();
    ui->bvsTable->horizontalHeader()->setStretchLastSection(true);
    //ui->bvsTable->resizeColumnsToContents();
    FloatDelegate *bvsDelegate = new FloatDelegate(ui->bvsTable);
    bvsDelegate->setDecimals(3);
    ui->bvsTable->setItemDelegateForColumn(3, bvsDelegate);
    ui->bvsTable->sortByColumn(0,Qt::AscendingOrder);
    connect(ui->bvsTable->selectionModel(),&QItemSelectionModel::currentRowChanged,
            this,&BondValenceDialog::bvsTable_currentRowChanged);

    binfoModel = new QStandardItemModel(this);
    QStringList binfoLabels = {"No.","Label","Type","x","y","z","Symm.Op.",
                               "Distance","Bond Valence"};
    binfoModel->setHorizontalHeaderLabels(binfoLabels);
    ui->bondsTable->verticalHeader()->setDefaultSectionSize(22);
    ui->bondsTable->setModel(binfoModel);
    ui->bondsTable->verticalHeader()->hide();    
    ui->bondsTable->resizeColumnsToContents();
    ui->bondsTable->horizontalHeader()->setStretchLastSection(true);
    FloatDelegate *binfoDelegate = new FloatDelegate(ui->bondsTable);
    binfoDelegate->setDecimals(5);
    ui->bondsTable->setItemDelegateForColumn(3, binfoDelegate);
    ui->bondsTable->setItemDelegateForColumn(4, binfoDelegate);
    ui->bondsTable->setItemDelegateForColumn(5, binfoDelegate);
    FloatDelegate *binfoDelegate1 = new FloatDelegate(ui->bondsTable);
    binfoDelegate1->setDecimals(3);
    ui->bondsTable->setItemDelegateForColumn(7, binfoDelegate1);
    ui->bondsTable->setItemDelegateForColumn(8, binfoDelegate1);
    ui->bondsTable->sortByColumn(0,Qt::AscendingOrder);

    ui->splitter->setStretchFactor(0,2);
    ui->splitter->setStretchFactor(1,1);

    mBvParDialog = new BvParDialog(this);
}

BondValenceDialog::~BondValenceDialog()
{
    delete ui;
}

int BondValenceDialog::on_bondvs(int whichModel)
{
    md = jav->GetModel(whichModel);
    wm = whichModel;    
    if (md->numSpec == 0) {
        QMessageBox::information(this, "Bond Valence",
                     "Program is unable to compute bond valences!");
        return 1;
    }
    int nats_au = md->nats_au;    
    selectedRow = 0;
    selectedRowPrevious = 0;

    float *bvs = new float[nats_au];
    int *cn = new int[nats_au];
    float gii;
    int cod = 1;
    jav_bond_valence_sum_(&whichModel, &nats_au, bvs, cn, &gii, &cod);

    ui->GIILabel->setText("GII: " + QString::number(gii,'f',3));

    ui->bvsTable->sortByColumn(0,Qt::AscendingOrder);

    QList<QStandardItem *> *bvInfo;
    bvInfo = md->FillBondValence(bvs,cn);

    delete[] bvs;
    delete[] cn;

    bvsModel->setRowCount(0);
    for(int row = 0; row < nats_au; row++)
        bvsModel->appendRow(bvInfo[row]);
    ui->bvsTable->selectRow(selectedRow);

    //test
//    QVector<QVector<int>> site;
//    md->FillSite(site);

    return 0;
}

void BondValenceDialog::bvsTable_currentRowChanged(const QModelIndex &current, const QModelIndex &previous)
{
    Q_UNUSED(previous);

    if (!current.isValid()) return;

    selectedRowPrevious = selectedRow;
    selectedRow = bvsModel->index(current.row(),0).data().toInt();
    int nrows = jav_bvs_ndinfo_(&wm,&selectedRow);    

    bond_info_t *dinfo = new bond_info_t[nrows];
    jav_bvs_dinfo_(&wm,&selectedRow,&nrows,dinfo);

    QList<QStandardItem *> *bvInfo;
    bvInfo = md->FillBondValenceInfo(nrows, dinfo);
    delete[] dinfo;

    binfoModel->setRowCount(0);
    for (int row = 0; row < nrows; row++){
        binfoModel->appendRow(bvInfo[row]);
    }
    ui->bondsTable->resizeColumnsToContents();
    ui->bondsTable->horizontalHeader()->setStretchLastSection(true);

    int sortedColumn = ui->bondsTable->horizontalHeader()->sortIndicatorSection();
    Qt::SortOrder order = ui->bondsTable->horizontalHeader()->sortIndicatorOrder();    
    ui->bondsTable->sortByColumn(sortedColumn,order);

    ui->atomLabel->setText("Bonds of atom " + bvsModel->index(current.row(),1).data().toString());

    //Unselect previuolsly selected atom, select atom selectedRow-1
    if (selectedRowPrevious > 0) md->atm[selectedRowPrevious-1].selected = false;
    md->atm[selectedRow-1].selected = true;
    md->atoms_sel.clear();
    md->atoms_sel.push_back(selectedRow);
    jav->Draw();
}

void BondValenceDialog::on_buttonBox_rejected()
{
    if (selectedRow > 0) {
        md->atm[selectedRow-1].selected = false;
        jav->Draw();
    }
    QDialog::reject();
}

void BondValenceDialog::on_exportButton_clicked()
{
    QSettings settings;
    QString selectedFilter("Text files (*.txt)");
    MainWindow *mainWin = getMainWindow();
    QString fileName = SaveDialog::run(this,
                                       tr("Export"),
                                       settings.value(DEFAULT_DIR_KEY).toString(),
                                       QFileInfo(mainWin->getCurrentFile()).baseName(),
                                       "Text files (*.txt);;All files (*)",
                                       "txt",
                                       selectedFilter);
    if (!fileName.isEmpty()) {
        settings.setValue(DEFAULT_DIR_KEY,QFileInfo(fileName).absolutePath());
        jav_export_bv_(fileName.toStdString().c_str(), fileName.length());
    }
}

void BondValenceDialog::on_pushButton_clicked()
{
    mBvParDialog->on_edit_bvs(md);
    if(mBvParDialog->exec() == QDialog::Accepted) {       
       int nats_au = md->nats_au;
       float *bvs = new float[nats_au];
       int *cn = new int[nats_au];
       float gii;
       int cod = 2;
       jav_bond_valence_sum_(&wm, &md->nats_au, bvs, cn, &gii, &cod);
       ui->GIILabel->setText("GII: " + QString::number(gii,'f',3));
       QList<QStandardItem *> *bvInfo;
       bvInfo = md->FillBondValence(bvs,cn);

       delete[] bvs;
       delete[] cn;

       bvsModel->setRowCount(0);
       for(int row = 0; row < nats_au; row++)
           bvsModel->appendRow(bvInfo[row]);

       int sortedColumn = ui->bvsTable->horizontalHeader()->sortIndicatorSection();
       Qt::SortOrder order = ui->bvsTable->horizontalHeader()->sortIndicatorOrder();
       ui->bvsTable->sortByColumn(sortedColumn,order);

       ui->bvsTable->selectRow(selectedRow);
    }
}
