#include "crystalsceneview.h"

#include <QKeyEvent>

CrystalSceneView::CrystalSceneView(QWidget *parent) :
    QOpenGLWidget(parent),
    camera(QVector3D(0.0f, 0.0f, 3.0))
{
    // tell keyboard handler to monitor certain keys
    m_keyboardMouseHandler.addRecognizedKey(Qt::Key_W);
    m_keyboardMouseHandler.addRecognizedKey(Qt::Key_S);
    m_keyboardMouseHandler.addRecognizedKey(Qt::Key_A);
    m_keyboardMouseHandler.addRecognizedKey(Qt::Key_D);
    m_keyboardMouseHandler.addRecognizedKey(Qt::Key_Escape);
}

CrystalSceneView::~CrystalSceneView()
{
    makeCurrent();
    shader.destroy();
}

void CrystalSceneView::initializeGL()
{
    initializeOpenGLFunctions();

    glEnable(GL_DEPTH_TEST);

    shader = ShaderProgram ("://shaders/default.vert", "://shaders/default.frag");
    cubes.create(shader);
    cubes.cubePositions << QVector3D( 0.0f,  0.0f,  0.0f)
                        << QVector3D( 2.0f,  5.0f, -15.0f)
                        << QVector3D(-1.5f, -2.2f, -2.5f)
                        << QVector3D(-3.8f, -2.0f, -12.3f)
                        << QVector3D( 2.4f, -0.4f, -3.5f)
                        << QVector3D(-1.7f,  3.0f, -7.5f)
                        << QVector3D( 1.3f, -2.0f, -2.5f)
                        << QVector3D( 1.5f,  2.0f, -2.5f)
                        << QVector3D( 1.5f,  0.2f, -1.5f)
                        << QVector3D(-1.3f,  1.0f, -1.5f);
}

void CrystalSceneView::paintGL()
{
    processInput();

    glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    shader.use();

    shader.setMat4("projection", projection);

    // camera/view transformation
    QMatrix4x4 view;
    view = camera.GetViewMatrix();
    shader.setMat4("view", view);

    cubes.render(shader);
}

void CrystalSceneView::resizeGL(int w, int h)
{
    projection.setToIdentity();
    projection.perspective(camera.Zoom, GLfloat(w) / h, 0.1f, 100.0f);
}

void CrystalSceneView::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Escape) {
        close();
    } else {
        m_keyboardMouseHandler.keyPressEvent(event);
        checkInput();
    }
}

void CrystalSceneView::keyReleaseEvent(QKeyEvent *event)
{
    m_keyboardMouseHandler.keyReleaseEvent(event);
    checkInput();
}

void CrystalSceneView::mousePressEvent(QMouseEvent *event)
{
    m_keyboardMouseHandler.mousePressEvent(event);
    checkInput();
}

void CrystalSceneView::mouseReleaseEvent(QMouseEvent *event)
{
    m_keyboardMouseHandler.mouseReleaseEvent(event);
    checkInput();
}

void CrystalSceneView::mouseMoveEvent(QMouseEvent *event)
{
    checkInput();
}

void CrystalSceneView::wheelEvent(QWheelEvent *event)
{
    m_keyboardMouseHandler.wheelEvent(event);
    checkInput();
}

void CrystalSceneView::checkInput()
{
    if (m_keyboardMouseHandler.keyDown(Qt::Key_W) ||
        m_keyboardMouseHandler.keyDown(Qt::Key_S) ||
        m_keyboardMouseHandler.keyDown(Qt::Key_A) ||
        m_keyboardMouseHandler.keyDown(Qt::Key_D))
    {
        m_inputEventReceived = true;
        update();
        return;
    }

    if (m_keyboardMouseHandler.buttonDown(Qt::LeftButton)) {
        // has the mouse been moved?
        if (m_keyboardMouseHandler.mouseDownPos() != QCursor::pos()) {
            m_inputEventReceived = true;
            update();
            return;
        }
    }

    if (m_keyboardMouseHandler.buttonDown(Qt::RightButton)) {
        // has the mouse been moved?
        if (m_keyboardMouseHandler.mouseDownPos() != QCursor::pos()) {
            m_inputEventReceived = true;
            update();
            return;
        }
    }

    // scroll-wheel turned?
    if (m_keyboardMouseHandler.wheelDelta() != 0) {
        m_inputEventReceived = true;
        update();
        return;
    }
}

void CrystalSceneView::processInput()
{
    if(!m_inputEventReceived) return;

    m_inputEventReceived = false;

    //qInfo() << "DELTA: " << deltaTime;
    deltaTime = 1.0;
    //const float cameraSpeed = 1.0; //deltaTime;
    if (m_keyboardMouseHandler.keyDown(Qt::Key_W))
    {
        //cameraPos += cameraSpeed * cameraFront;
        camera.ProcessKeyboard(FORWARD, deltaTime);
        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }
    if (m_keyboardMouseHandler.keyDown(Qt::Key_S))
    {
        //cameraPos -= cameraSpeed * cameraFront;
        camera.ProcessKeyboard(BACKWARD, deltaTime);
        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }
    if (m_keyboardMouseHandler.keyDown(Qt::Key_A))
    {
        //cameraPos -= QVector3D::crossProduct(cameraFront, cameraUp).normalized() * cameraSpeed;
        camera.ProcessKeyboard(LEFT, deltaTime);
        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }
    if (m_keyboardMouseHandler.keyDown(Qt::Key_D))
    {
        //cameraPos += QVector3D::crossProduct(cameraFront, cameraUp).normalized() * cameraSpeed;
        camera.ProcessKeyboard(RIGHT, deltaTime);
        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }

    if (m_keyboardMouseHandler.buttonDown(Qt::LeftButton)) {

        // Handle rotations
        // get and reset mouse delta (pass current mouse cursor position)
        QPoint mouseDelta = m_keyboardMouseHandler.resetMouseDelta(QCursor::pos()); // resets the internal position
        float xoffset = static_cast<float>(mouseDelta.x());
        float yoffset = -static_cast<float>(mouseDelta.y()); // reversed since y-coordinates go from bottom to top

        camera.ProcessMouseMovement(xoffset, yoffset);

        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }

    if (m_keyboardMouseHandler.buttonDown(Qt::RightButton)) {
        QPoint mouseDelta = m_keyboardMouseHandler.resetMouseDelta(QCursor::pos());
        //qInfo() << "Mouse delta: " << mouseDelta.x() << mouseDelta.y();

        //FIX this action

        // QVector3D translation(mouseDelta.x(), -mouseDelta.y(), 0.0f);
        // qInfo() << "TRANSL: " << translation.normalized();
        // qInfo() << "CAMERA: " << cameraPos;
        // cameraPos += translation.normalized() * 0.05; //QVector3D(newPos.normalized().x(), newPos.normalized().y(), 0);
        // //cameraFront += translation.normalized() * 0.05;
        m_keyboardMouseHandler.clearWasPressedKeyStates();
    }

    int wheelDelta = m_keyboardMouseHandler.resetWheelDelta();
    if (wheelDelta != 0) {
        qInfo() << "Wheel: " << wheelDelta;
        camera.ProcessMouseScroll(static_cast<float>(wheelDelta));
        resizeGL(width(), height());
    }
}
