#include "grafico.h"
#include <QtMath>
#include <QFontMetrics>
#include <QFontDatabase>
#include <QGraphicsTextItem>
#include <QGraphicsProxyWidget>
#include <QLabel>
#include <QLineF>
#include <QGraphicsSceneMouseEvent>
#include "qt_utils.h"

#define SEC_TICS 5
#define EPS 1e-6

namespace GraphicSpace 
{

    Plot::Plot(QGraphicsScene *parent, int n, float x[], float y[])
    {
        m_Scene = parent;
        if(!points.empty())
            points.clear();
        if(!ipoints.empty())
            ipoints.clear();
        ifLines = true;
        ifMarkers = false;
        ifRefl = false;
        QColor blue("#FF0000E5");
        QColor violet("#FFB300B3");
        color_line = blue;
        color_marker = blue;
        color_refl = violet;

        numPoints = n;
        for(int i=0; i<n; i++)
        {
            QPointF p;
            p.setX(x[i]);
            if(y != NULL)
               p.setY(y[i]);
            else
            {
               p.setY(0);
               ifRefl = true;
            }
            points.push_back(p);
        }
        xmin = (*min_element(points.begin(), points.end(), Cfrx)).x();
        xmax = (*max_element(points.begin(), points.end(), Cfrx)).x();
        ymin = (*min_element(points.begin(), points.end(), Cfry)).y();
        ymax = (*max_element(points.begin(), points.end(), Cfry)).y();
    }

    void Plot::GetLim(float *x1, float *x2, float *y1, float *y2)
    {
        *x1 = xmin;
        *x2 = xmax;
        *y1 = ymin;
        *y2 = ymax;
    }

    void Plot::DisegnaRiflessi(int bottom)
    {
        Grafico *scene = dynamic_cast<Grafico *> (m_Scene);

        std::vector<QLineF> ss;
        for(int i=0; i<numPoints; i++)
        {
            if(InLimits(ipoints.at(i).x()))
            {
               QLineF s(ipoints.at(i).x(), bottom+5, ipoints.at(i).x(), bottom+15);
               ss.push_back(s);
            }
        }
        if(ss.empty())
           return;
        int nss = (int)ss.size();
        for(int i=0; i<nss; i++)
            scene->addLine(ss.at(i), QPen(color_refl));
        ss.clear();
    }

    bool Plot::InLimits(QPoint p1)
    {
        QPointF pp1, pp2;
        Grafico *scene = dynamic_cast<Grafico *> (m_Scene);
        scene->GetLim(&pp1, &pp2);
        QPoint minlim = scene->ToScreen(pp1);
        QPoint maxlim = scene->ToScreen(pp2);
        return ((p1.x() >= minlim.x()) && (p1.x() <= maxlim.x()) );
    }

    bool Plot::InLimits(int x1)
    {
        Grafico *scene = dynamic_cast<Grafico *> (m_Scene);
        QPointF pp1, pp2;
        scene->GetLim(&pp1, &pp2);
        QPoint minlim = scene->ToScreen(pp1);
        QPoint maxlim = scene->ToScreen(pp2);
        return ((x1 >= minlim.x()) && (x1 <= maxlim.x()));
    }

    void Plot::setColor_line(const QColor &value)
    {
        color_line = value;
    }

    void Plot::Disegna()
    {
        Grafico *scene = dynamic_cast<Grafico *> (m_Scene);
        std::vector<QPoint> pps;
        for(int i=0; i<numPoints; i++)
        {
            if(InLimits(ipoints.at(i)))
            {
               QPoint pp(ipoints.at(i));
               pps.push_back(pp);
            }
        }
        if(pps.empty())
           return;
        int npp=(int) pps.size();
        for(int i=0; i<npp-1; i++)
        {
            //scene->addLine(QLineF(pps.at(i), pps.at(i+1)), QPen(color_line));
            AntialiasedLine *line = new AntialiasedLine(pps.at(i), pps.at(i+1), color_line);
            scene->addItem( line);
        }
        pps.clear();
    }

void CalcolaTics(float vmin, float vmax, double *step_t, int *ntics,
                  double *primotic, int *disp, int *nci)
{
   int    espo, i1, i2;
   double delta, step_tic, pristep, primopri, secstep, primosec;

   delta = (double) (vmax - vmin);
   if(abs(delta) < EPS)
   {
      //g_print("Warning: vmin = vmax\n");
      delta = 10;
   }
   espo = 0;
   while(((int) delta) >= 10)
   {
       delta /= 10.;
       espo--;
   }
   while(((int) delta) == 0)
   {
       delta *= 10.;
       espo++;
   }
   if(delta <= 1.5)
      step_tic = 0.1/SEC_TICS;
   else if(delta <= 3.)
      step_tic = 0.2/SEC_TICS;
   else if(delta <= 7.)
      step_tic = 0.5/SEC_TICS;
   else 
      step_tic = 1./SEC_TICS;

   step_tic *= 2;

   *ntics  = (int)(delta / step_tic) + 1;
   *step_t = step_tic / qPow(10., espo);
   secstep = *step_t;
   pristep = step_tic * SEC_TICS / qPow(10., espo);

   i1 = (int)(vmin*qPow(10., (espo+1))+0.5);
   i2 = (int)(pristep*qPow(10., (espo+1))+0.5);
   while((i1 % i2) != 0)
      i1++;
   primopri = (double)i1 / qPow(10., (espo+1));
   if(primopri < vmin)
      primopri += pristep;

   i1 = (int)(SEC_TICS*vmin*qPow(10.,(espo+1))+0.5);
   i2 = (int)(SEC_TICS*secstep*qPow(10., (espo+1))+0.5);
   while((i1 % i2) != 0)
      i1++;
   primosec = (double)i1 / (SEC_TICS * qPow(10., (espo+1)));
   if(primosec < vmin)
   {
      primosec += secstep;
      (*ntics)--;
   }
   *primotic = primosec;
   *disp = (int)((primopri - primosec)/secstep);

   delta = pristep;
   espo = 0;
   while(((int) delta) >= 10)
   {
       delta /= 10.;
       espo--;
   }
   while(((int) delta) == 0)
   {
       delta *= 10.;
       espo++;
   }
   if(espo > 0)
      *nci = espo+1;
   else
      *nci = espo;
}

     Grafico::Grafico(QStatusBar *bar, QPushButton *button, QObject *parent)
        :QGraphicsScene(parent)
     {
         m_bar = bar;
         resetButton = button;
         if(!Plots.empty())
            Plots.clear();
         QColor black("#FF333333");
         color_axes = black;
         //Gfont = QFont("Monospace");
         //Gfont.setStyleHint(QFont::TypeWriter);
         Gfont = QFont(QFontDatabase::systemFont(QFontDatabase::FixedFont));
         setFont(Gfont);
         QFontMetrics fm(Gfont);
         wfont = fm.maxWidth();
         hfont = fm.height();
         xmin = xmax = ymin = ymax = 0;
         start_move = false;
         Xiniz = Yiniz = -1;
         Wrect = Hrect = 0;
     }
    
     void Grafico::ClearScene()
     {
         clear();
         if(!Plots.empty())
             Plots.clear();
     }

     void Grafico::SetArea(int w, int h)
     {
         Width  = w;
         Height = h;
     }

     void Grafico::SetLim(float x1, float x2, float y1, float y2, bool force)
     {
         xmin = x1;
         xmax = x2;
         if(force)
         {
            ymin = y1;
            ymax = y2;
         }
         else
         {
            ymin = min(ymin, y1);
            ymax = max(ymax, y2);
         }
     }

     void Grafico::CalcolaScala()
     {
         int dispx, dispy, ncix, nciy, xtics, ytics;
         double tic1x, stepx, tic1y, stepy;

         CalcolaTics(xmin, xmax, &stepx, &xtics, &tic1x, &dispx, &ncix);
         CalcolaTics(ymin, ymax, &stepy, &ytics, &tic1y, &dispy, &nciy);

         right = left = top = bottom = hfont;
         bottom += 4*hfont;
         left += (abs(nciy)+6)*wfont+hfont;//+10)*wfont;
         top    += 2*hfont;
         XScale = (float)(Width - left - right)/(xmax -xmin);
         YScale = (float)(Height - top - bottom)/(ymax -ymin);
         dx = (float) left / XScale - xmin;
         dy = (float) top / YScale + ymax;
         assex.valmin = xmin;
         assex.valmax = xmax;
         assex.bottom = bottom;
         assex.top    = top;
         assex.right  = right;
         assex.left   = left;
         assex.deltax = Width-right;
         assex.deltay = Height-bottom;
         assex.scale  = XScale;
         assex.start  = tic1x;
         assex.step   = stepx;
         assex.tics   = xtics;
         assex.fprim  = dispx;
         assex.fsec   = (int) ((tic1x - xmin) * XScale) + left;
         assex.ori    = 'h';
         assex.label  = labelx;
         assex.ncifre = ncix;
         assex.ifSect = true;

         assey.valmin = ymin;
         assey.valmax = ymax;
         assey.bottom = bottom;
         assey.left   = left;
         assey.top    = top;
         assey.right  = right;
         assey.deltax = Width-right;
         assey.deltay = Height-bottom;
         assey.scale  = YScale;
         assey.start  = tic1y;
         assey.step   = stepy;
         assey.tics   = ytics;
         assey.fprim  = dispy;
         assey.fsec   = (int) ((tic1y - ymin) * YScale) + top;
         assey.ori    = 'v';
         assey.label  = labely;
         assey.ncifre = nciy;
         assey.ifSect = true;

         labx.setX((assex.deltax+assex.left)/2);
         labx.setY(assex.deltay+(int)(2.5*hfont));
         if(nciy > 0)
            laby.setX(4);
         else
            laby.setX((int)(hfont*1.1));
         laby.setY(assey.deltay/2+wfont*5);
     }

     QPointF Grafico::FromScreen(QPoint p)
     {
         QPointF p1 = {(float)p.x() / XScale - dx,
                      dy - (float)p.y() / YScale};
         return p1;
     }

     QPoint Grafico::ToScreen(QPointF p)
     {
         QPoint p1 = {(int)((p.x() + dx) * XScale),
                       (int)((dy - p.y()) * YScale)};
         return p1;
     }

     void Grafico::DrawText(QFont f, QString s, QColor c, QPoint pos)
     {
         //QGraphicsTextItem *text = addText(s);
         //text->setFont(f);
         //text->setDefaultTextColor(c);
         //text->setPos(pos);
         QLabel *label = new QLabel(s);
         QString styleSheet = LabelStyle(QColor("#FFFFFF"), c);
         label->setFont(f);
         label->setStyleSheet(styleSheet);

         QGraphicsProxyWidget *proxy = addWidget(label);
         proxy->setPos(pos);
     }

     void Grafico::DrawVText(QFont f, QString s, QColor c, QPoint pos)
     {
         QLabel *label = new QLabel(s);
         QString styleSheet = LabelStyle(QColor("#FFFFFF"), c);
         label->setFont(f);
         label->setStyleSheet(styleSheet);

         QGraphicsProxyWidget *proxy = addWidget(label);
         proxy->setPos(pos);
         proxy->setRotation(270);
     }

     void Grafico::DrawAxes(Axes axe)
     {
          int k, x0, x1, y0, y1, ntics;

          switch(axe.ori)
          {
          case 'h':
                if(!axe.label.isEmpty())
                   DrawText(Gfont, axe.label, color_axes, labx);
                
                x0 = axe.left;
                x1 = axe.deltax;
                y0 = axe.deltay;
                y1 = y0;
                break;
          case 'H':
                x0 = axe.left;
                x1 = axe.deltax;
                y0 = axe.top;
                y1 = y0;
                break;
          case 'v':
                if(!axe.label.isEmpty())
                   DrawVText(Gfont, axe.label, color_axes, laby);
                x0 = axe.left;
                x1 = x0;
                y0 = axe.top; 
                y1 = axe.deltay;
                break;
          case 'V':
                x0 = axe.deltax;
                x1 = x0;
                y0 = axe.top; 
                y1 = axe.deltay;
                break;
          }
          addLine(QLine(x0,y0, x1, y1), QPen(color_axes));
          ntics = axe.tics;
          QLine *segs = new QLine [ntics]();
          vector<QLine> Segs;
          for(int i=0; i<ntics; i++)
          {
              int x1, y1;
              switch(axe.ori)
              {
              case 'h':
                    k = 1;
                    x1 = axe.fsec + (int)((axe.step*(double)i)*axe.scale);
                    if(((i-axe.fprim) % SEC_TICS) == 0)
                    {
                        k = 2;
                        float duetheta = (float) (axe.start+i*axe.step);
                        int nc = max(0, axe.ncifre);
                        QString dt = (QString::number(duetheta, 'f', nc)).trimmed();
                        int lu = wfont*(int)dt.length();
                        DrawText(Gfont, dt, color_axes, QPoint(x1-lu, axe.deltay+hfont));
                    }
                    segs[i] = QLine(x1, axe.deltay, x1, axe.deltay+5*k);
                    if(k == 2)
                       Segs.push_back(segs[i]);
                    break;
              case 'v':
                    k = 1;
                    y1 = axe.top + axe.deltay - axe.fsec
                                  - (int)((axe.step*(double)i)*axe.scale);
                    if(((i-axe.fprim) % SEC_TICS) == 0)
                    {
                        k = 2;
                        int nc = max(0, axe.ncifre);
                        QString dt = (QString::number(axe.start+i*axe.step, 'f', nc)).trimmed();
                        //if(nc)
                           DrawText(Gfont, dt, color_axes, QPoint(axe.left-5*wfont, y1-hfont/2));
                        //else
                            //DrawText(Gfont, dt, color_axes, QPoint(axe.left-3*wfont, y1-hfont/2));
                    }
                    segs[i] = QLine(axe.left - 5 * k, y1, axe.left, y1);
                    if(k == 2)
                       Segs.push_back(segs[i]);
                    break;
              }
          }
          if(axe.ifSect)
          {
              for(int j=0; j<ntics; j++)
                  addLine(segs[j], QPen(color_axes));
          }
          else
          {
            for(size_t i=0; i<Segs.size(); i++)
                addLine(Segs.at(i), QPen(color_axes));
          }
          delete [] segs;
          Segs.clear();
     }

     void Grafico::AddPlot(float x[], float y[], int num, float tmin,
                           float tmax, bool uset)
     {
         float x1, y1, x2, y2;
         Plot plotn = Plot(this, num, x, y);
         plotn.GetLim(&x1, &x2, &y1, &y2);
         if(!uset)
         {
            tmin = x1;
            tmax = x2;
         }
         SetLim(tmin, tmax, y1, y2);
         xmin0 = tmin;
         xmax0 = tmax;
         if(y != NULL)
         {
            ymin0 = y1;
            ymax0 = y2;
         }
         Plots.push_back(plotn);
     }

     void Grafico::Disegna()
     {
         for(size_t i=0; i<Plots.size(); i++)
         {
             Plots[i].ToScreen();
             if(!Plots[i].IsRefl())
                Plots[i].Disegna();
         }
         QPointF pp1, pp2;
         GetLim(&pp1, &pp2);
         QPoint minlim = ToScreen(pp1);
         QPoint maxlim = ToScreen(pp2);
         ClearArea(minlim.x(), 0, maxlim.x()-minlim.x(), maxlim.y());
         ClearArea(minlim.x(), minlim.y(), maxlim.x()-minlim.x(), Height-minlim.y());
         for(size_t i=0; i<Plots.size(); i++)
         {
             if(Plots[i].IsRefl())
                Plots[i].DisegnaRiflessi(assey.deltay);
         }

         DrawAxes(assex);
         DrawAxes(assey);

         assex.ori = 'H';
         assey.ori = 'V';
         DrawAxes(assex);
         DrawAxes(assey);
         assex.ori = 'h';
         assey.ori = 'v';

     }

     void Grafico::ClearArea(int x, int y, int w, int h) 
     {
         setBackgroundBrush(QBrush("#FFFFFF"));
         addRect(x,y,x+w-1,y+h-1, QPen("#FFFFFF"), QBrush("#FFFFFF"));
     }

     void Grafico::Aggiorna()
     {
          clear();
          CalcolaScala();
          Disegna();
     }

     void Plot::ToScreen()
     {
         Grafico *scene = dynamic_cast<Grafico *> (m_Scene);
         if(!ipoints.empty())
            ipoints.clear();
         for(int i=0; i<numPoints; i++)
             ipoints.push_back(scene->ToScreen(points.at(i)));
     }


     void Grafico::DrawDashRect (int ix, int iy, int w, int h)
     {
        QPen pen("#000000");

        for(int x=ix; x<ix+w; x+=10)
        {
            addLine(QLine(x, iy, x+3, iy), pen);
            addLine(QLine(x, iy+h-1, x+3, iy+h-1), pen);
        }

        for(int y=iy; y<iy+h; y+=10)
        {
            addLine(QLine(ix, y, ix, y+3), pen);
            addLine(QLine(ix+w-1, y, ix+w-1, y+3), pen);
        }
     }

     void Grafico::GetLim(QPointF *p1, QPointF *p2)
     {
         p1->setX(xmin);
         p1->setY(ymin);
         p2->setX(xmax);
         p2->setY(ymax);

     }

     void Grafico::ResetLim()
     {
         xmin = xmin0;
         xmax = xmax0;
         ymin = ymin0;
         ymax = ymax0;
     }

     void Grafico::sizeChanged(int w, int h)
     {
         SetArea(w, h);
         setSceneRect(0, 0, w, h);
         Aggiorna();
     }

     int Grafico::numPlots()
     {
         return Plots.size();
     }

     void Grafico::setColorPlot(int idPlot, const QColor &colorLine)
     {
         Plots[idPlot].setColor_line(colorLine);
     }

     void Grafico::MouseMove(int x, int y)
     {
         if(start_move)
         {
            if (Xiniz < 0 )
            {
                Xiniz = x;
                Yiniz = y;
            }
            else
            {
                Wrect = x - Xiniz + 1;
                Hrect = y - Yiniz + 1;
                QPen pen("#444444");
                pen.setStyle(Qt::DashLine);
                QBrush transp("#00FFFFFF");
                QGraphicsItem *m_dashedRect = itemAt(Xiniz, Yiniz, QTransform());
                if(m_dashedRect)
                    removeItem(m_dashedRect);
                m_dashedRect = addRect(min(Xiniz, x), min(Yiniz, y), abs(Wrect), abs(Hrect),
                        pen, transp);
            }
         }
     }

     void Grafico::mouseReleaseEvent(QGraphicsSceneMouseEvent *)
     {
         if(start_move)
         {
            if((Wrect * Hrect) != 0)
            {
                QPoint ip1(min(Xiniz, Xiniz+Wrect), max(Yiniz, Yiniz+Hrect));
                QPoint ip2(max(Xiniz, Xiniz+Wrect), min(Yiniz, Yiniz+Hrect));
                QPointF p1 = FromScreen(ip1);
                QPointF p2 = FromScreen(ip2);
                QPointF pp1, pp2;
                GetLim(&pp1, &pp2);
                float x1=pp1.x(), y1=pp1.y(), x2=pp2.x(), y2=pp2.y();
                if(x1 < p1.x())
                   x1 = p1.x();
                if(x1 > p2.x())
                   x1 = p2.x();
                if(x2 < p1.x())
                   x2 = p1.x();
                if(x2 > p2.x())
                   x2 = p2.x();
                if(y1 < p1.y())
                   y1 = p1.y();
                if(y1 > p2.y())
                   y1 = p2.y();
                if(y2 < p1.y())
                   y2 = p1.y();
                if(y2 > p2.y())
                   y2 = p2.y();
                if((abs(y1-y2) > 1e-8) && (abs(x1-x2) >1e-8))
                {
                    SetLim(x1, x2, y1, y2, true);
                    resetButton->setEnabled(true);
                }
                Aggiorna();
            }
            Xiniz = -1;
            Yiniz = -1;
            Wrect = 0;
            Hrect = 0;
            start_move = false;
            return;
         }
     }

     void Grafico::mousePressEvent(QGraphicsSceneMouseEvent * mouseEvent)
     {
         if(mouseEvent->buttons() & Qt::LeftButton)
            start_move = true;
     }

     void Grafico::mouseMoveEvent(QGraphicsSceneMouseEvent * mouseEvent)
     {
          QPoint ip((int) mouseEvent->scenePos().x(), (int) mouseEvent->scenePos().y());
          Qt::MouseButton button = mouseEvent->button();

          if(mouseEvent->buttons() & Qt::LeftButton)
              MouseMove(ip.x(), ip.y());
          else if(button == Qt::NoButton)
          {
              QPointF p = FromScreen(ip);
              QString stringa = mMessage(p);
              m_bar->showMessage(stringa);
          }
     }

     void Grafico::ResetZoom()
     {
         ResetLim();
         Aggiorna();
     }

}               // end namespace
