#include "mopacdialog.h"
#include "ui_mopacdialog.h"
#include "javkeysettings.h"
#include "qt_utils.h"
#if defined(JAV_WIDGET)
#include "crystalviewwidget.h"
#else
#include "javwindow.h"
#endif
#include "properties.h"
#include "fileutils.h"
#include "mainwindow.h"

#include <QFileDialog>
#include <QSettings>
#include <QStandardPaths>
#include <QTextStream>
#include <QProcess>
#include <QMessageBox>
#include <QFileInfo>
#include <QDebug>

typedef float coord_type[3];
typedef int   conn_type[3];

bool Importa(string nomefile, int whichModel, mat3x3 rmI);
extern "C" void cartesiantozmatrix_forc_(atom_type *atm, int *n, float I_coor[][3], int conn[][3], int *wm);
extern "C" void jav_optimize_geometry_mopac_(int *code, int *whichModel, atom_type *kptra, int *nptra);

MopacDialog::MopacDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::MopacDialog)
{
#if defined(JAV_WIDGET)
    jav = qobject_cast<CrystalViewWidget*>(parent);
#else
    jav = qobject_cast<JavWindow*>(parent);
#endif
    ui->setupUi(this);

    //This line is important to obtain a correct resize of window after hiding
    this->layout()->setSizeConstraint(QLayout::SetNoConstraint);

    //Fix size to avoid resize of button Show/Hide Preview
    const QSize  buttonSize(120,25);
    ui->previewButton->setMinimumSize(buttonSize);

    ui->buttonBox->button(QDialogButtonBox::Apply)->setText("Submit");
    ui->buttonBox->button(QDialogButtonBox::Apply)->setDefault(true);
    ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setText("Use Form");
    ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(false);

    on_previewButton_clicked(); // hide Preview

    connect(ui->formatCombo,SIGNAL(currentIndexChanged(int)),this,SLOT(updatePreview(int)));
    connect(ui->calculationCombo,SIGNAL(currentIndexChanged(int)),this,SLOT(updatePreview(int)));
    connect(ui->methodCombo,SIGNAL(currentIndexChanged(int)),this,SLOT(updatePreview(int)));
    connect(ui->multiplicityCombo,SIGNAL(currentIndexChanged(int)),this,SLOT(updatePreview(int)));
    connect(ui->chargeSpinBox,SIGNAL(valueChanged(int)),this,SLOT(updatePreview(int)));
    connect(ui->titleLineEdit,SIGNAL(textChanged(QString)),this,SLOT(titleChanged(QString)));
    connect(ui->textEdit,&QTextEdit::textChanged,this,&MopacDialog::previewChanged);
}

MopacDialog::~MopacDialog()
{
    delete ui;
}

void MopacDialog::on_previewButton_clicked()
{
    int width = this->size().width();

    if (ui->textEdit->isHidden()) {
        ui->textEdit->show();
        ui->previewButton->setText("Hide Preview");
        FillPreview();
    } else {
        ui->textEdit->hide();
        ui->previewButton->setText("Show Preview");
    }

    //Resize only in height
    QSize size = sizeHint();
    size.setWidth(width);
    resize(size);
}

void MopacDialog::on_programButton_clicked()
{    
    QString fileName = QFileDialog::getOpenFileName(this, tr("Select Executable File"),
                                                    QDir::homePath(),
#ifdef Q_OS_WIN32
                                                    tr("Program Files (*.exe);;All files (*)"));
#else
                                                    tr("All files (*);;Program Files (*.exe)"));
#endif
    if (!fileName.isEmpty()) {
        ui->programCombo->insertItem(0,QDir::toNativeSeparators(fileName));
        ui->programCombo->setCurrentIndex(0);        
    }
}

void MopacDialog::openMOPAC(int whichModel)
{
    md = jav->GetModel(whichModel);
    wm = whichModel;
    outFileName.clear();
    if (ui->programCombo->currentText().isEmpty()) findMOPAC();
    FillPreview();
    int code = 0;
    int nptra = 0;
    atom_type *kptr = NULL;
    jav_optimize_geometry_mopac_(&code,&wm,kptr,&nptra);
}

void MopacDialog::findMOPAC()
{
    //Find in configuration file
    QSettings settings;
    if(settings.contains(MOPAC_EXE)) {
       ui->programCombo->addItem(settings.value(MOPAC_EXE).toString());
       return;
    }

    //Find in PATH
    QStringList files;
    fileutils::findFiles(files,"mopac*.exe");
#ifdef Q_OS_LINUX
    if (files.isEmpty()) {
       //executable for version 22 on Linux has name MOPAC
       fileutils::findFiles(files,"mopac");
    }
#endif
    if (files.isEmpty()) {

    //Find in standard location
#ifdef Q_OS_WIN32
        //QString dirMopac = qEnvironmentVariable("PROGRAMFILES(x86)") + "/MOPAC/";
        QString dirMopac = qEnvironmentVariable("PROGRAMFILES") + "/MOPAC/";
#else
        QString dirMopac = "/opt/mopac/";
#endif
        fileutils::findFiles(files,dirMopac,"mopac*.exe");
#ifdef Q_OS_LINUX
        if (files.isEmpty()) {
            //executable for version 22 on Linux has name MOPAC
            fileutils::findFiles(files,dirMopac,"mopac");
        }
#endif
    }

    // revert order to have mopac.exe as first
    for (int i = files.size()-1; i >= 0; i--) {
        ui->programCombo->addItem(QDir::toNativeSeparators(files.at(i)));
    }
}

void MopacDialog::on_buttonBox_clicked(QAbstractButton *button)
{
    QDialogButtonBox::StandardButton stdButton = ui->buttonBox->standardButton(button);

    if(stdButton == QDialogButtonBox::Apply) {   // Submit
       QSettings settings;
       QString programLocation = ui->programCombo->currentText();
       if (!programLocation.isEmpty()) settings.setValue(MOPAC_EXE,programLocation);

       //Set working directory
       QString wDir;
       MainWindow *mainWindow = getMainWindow();
//       if (!jav->getCurrentFilePath().isEmpty()) {
       if (!mainWindow->getCurrentFile().isEmpty()) {
           wDir = QFileInfo(mainWindow->getCurrentFile()).absolutePath();
       } else {
           wDir = QDir::tempPath();
       }

       //Save Preview in inFileName
       QString inFileName = QString(md->modname.c_str()).append("_mopac.mop");
       QFile mopacFile(wDir + "/" + inFileName);
       if (mopacFile.open(QIODevice::WriteOnly | QIODevice::Text)) {
           QTextStream stream(&mopacFile);
           stream << ui->textEdit->toPlainText();
           mopacFile.close();
       }

       //Run MOPAC
       QProcess *process = new QProcess(this);
       QStringList arguments;       
       arguments << inFileName;
       process->setProcessChannelMode(QProcess::MergedChannels);
       process->setWorkingDirectory(wDir);
       QDialog::accept();
       process->start(programLocation,arguments);
       jav->waitProcess("Running MOPAC ...");
       if (!process->waitForFinished()) {
           jav->endProcess();
           jav->setStatusBarText("Program ends in error");
           QMessageBox::critical(this,tr("MOPAC error"),
                                 tr("MOPAC program failed.\n%1").arg(process->errorString()));           
       } else {
           jav->endProcess();
           QString ss(process->readAll());
           if (!ss.contains("normally")) {
               jav->setStatusBarText("Program ends in error");
               QMessageBox::critical(this,tr("MOPAC error"),ss);
           } else {
               jav->setStatusBarText("Running MOPAC ... done!");
               outFileName = wDir + "/" + QString(md->modname.c_str());
               outFileName.append("_mopac.out");
               bool result = Importa(outFileName.toStdString(), wm, Properties::omI);
               if (!result) {
                   QMessageBox::critical(this,tr("MOPAC error"),
                                         tr("MOPAC genereted error.\n"
                                            "Read the end of the output file:\n%1").arg(
                                             outFileName));
               }
           }
       }
       jav->clearStatusBarText();

    } else if (stdButton == QDialogButtonBox::Cancel) {

    } else if (stdButton == QDialogButtonBox::Reset) {
        ui->calculationCombo->setCurrentIndex(1);
        ui->methodCombo->setCurrentIndex(5);
        ui->formatCombo->setCurrentIndex(0);
        FillPreview();
        this->setEnabled(true);
        ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(false);
    } else if (stdButton == QDialogButtonBox::RestoreDefaults) {  //Use Form
        int ret = QMessageBox::question(this, "Use Form",
                  "Would you like to update the preview text, losing all changes made in the MOPAC input deck preview pane?");
        if (ret == QMessageBox::Yes) {
            FillPreview();
            ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(false);
        }
    }
}

void MopacDialog::updatePreview(int index)
{    
    Q_UNUSED(index);
    FillPreview();    
}

void MopacDialog::titleChanged(QString text)
{
    Q_UNUSED(text);
    FillPreview();
}

void MopacDialog::previewChanged()
{    
    this->setEnabled(false);
    ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(true);
}

void MopacDialog::setEnabled(bool state)
{
    ui->titleLineEdit->setEnabled(state);
    ui->titleLabel->setEnabled(state);
    ui->calculationLabel->setEnabled(state);
    ui->calculationCombo->setEnabled(state);
    ui->chargeLabel->setEnabled(state);
    ui->chargeSpinBox->setEnabled(state);
    ui->formatLabel->setEnabled(state);
    ui->formatCombo->setEnabled(state);
    ui->methodLabel->setEnabled(state);
    ui->methodCombo->setEnabled(state);
    ui->multiplicityLabel->setEnabled(state);
    ui->multiplicityCombo->setEnabled(state);
}

QString MopacDialog::getOutFileName() const
{
    return outFileName;
}

void MopacDialog::FillPreview()
{
    QString str;
    QTextStream strm(&str);
    strm << "AUX LARGE "
         << "CHARGE=" << ui->chargeSpinBox->text().toInt() << " "
         << ui->multiplicityCombo->currentText().toUpper() << " ";
    if (ui->calculationCombo->currentIndex() == 0)
        strm << "NOOPT ";
    else if (ui->calculationCombo->currentIndex() == 2)
        strm << "NOOPT OPT-H ";
    strm << ui->methodCombo->currentText().remove(QChar('-')).toUpper() << "\n";
    strm << ui->titleLineEdit->text() << "\n";
    strm << "\n";    
    if (ui->formatCombo->currentIndex() == 0) {
        md->FillPreview(strm);
    } else {
        atom_type *atml = md->FromJavatomToAtom();
        coord_type *I_coor = new coord_type[md->nats_au]();
        conn_type *conn = new conn_type[md->nats_au]();
        int km = wm++;
        cartesiantozmatrix_forc_(atml, &md->nats_au, I_coor, conn, &km);
        md->FillPreview(strm, &I_coor[0], &conn[0]);
        delete [] I_coor;
        delete [] conn;
        delete [] atml;
    }

    //Set text avoiding textChanged signal
    ui->textEdit->blockSignals(true);
    ui->textEdit->setText(str);
    ui->textEdit->blockSignals(false);
}
