#include "rietvelddialog.h"
#include "ui_rietvelddialog.h"
#include "mainwindow.h"
#include "phasewidget.h"
#include "qt_utils.h"
#include "crystaldata.h"

#include <QDebug>

extern "C" void get_dataset(dataset_t *data);
extern "C" void set_dataset(dataset_t *data);
extern "C" void get_back_settings(BackgroundSettings *backSettings);
extern "C" void set_refcond(refine_condition_type *rcond);
extern "C" void finishriet_glob(refine_condition_type *rcond, dataset_t *data);
extern "C" void c_applyriet(int modzero, int modcell[], int modwei, int modfun[], int modpo[]);
extern "C" void c_exitriet();
extern "C" int crystal_number();

const int firstPhaseTab = 2;

RietveldDialog::RietveldDialog(MainWindow *mw) :
    QDialog(mw),
    ui(new Ui::RietveldDialog),
    mWindow(mw),
    wSchemeChanged(false),
    numPhase(0)
{
    ui->setupUi(this);
    ui->buttonBox->button(QDialogButtonBox::Apply)->setText("Refine");
    ui->buttonBox->button(QDialogButtonBox::Apply)->setIcon(QIcon(":/pixmaps/pixmaps/runconfig.png"));

    //connect(ui->autoProfCheckBox,&QCheckBox::clicked,this,&RietveldDialog::nCycleSpinEnabled);
    //connect(ui->autoStructCheckBox,&QCheckBox::clicked,this,&RietveldDialog::nCycleSpinEnabled);
    connect(ui->autoProfCheckBox,&QCheckBox::clicked,this,&RietveldDialog::autoProfileClicked);
    connect(ui->autoStructCheckBox,&QCheckBox::clicked,this,&RietveldDialog::autoProfileClicked);
    connect(this, &RietveldDialog::rejected, [this](){
        mWindow->clearStatusMessage();
        mWindow->restoreEnabledActions();
    });
    connect(ui->weightSchemeComboBox, QOverload<int>::of(&QComboBox::currentIndexChanged), this, [=](int index){
       wSchemeChanged = (rietCond.typewei != index);
    });
    connect(ui->lebailCheckBox, &QCheckBox::clicked, this, [=](bool state){
        qInfo() << "Checkbox state:" << state;
        setLeBailRefineEnabled(!state);
    });
    connect(ui->buttonBox, &QDialogButtonBox::clicked, this, &RietveldDialog::onButtonBoxClicked);
}

RietveldDialog::~RietveldDialog()
{
    delete ui;
}

void RietveldDialog::waitSignalIsEmitted()
{
    bool isWaitCursor = stopWaitCursor();
    QEventLoop loop;
    connect(this,&QDialog::finished,&loop,&QEventLoop::quit);
    loop.exec();
    if (isWaitCursor) QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
}

void RietveldDialog::setDialog(const refine_condition_type &rcond)
{
    rietCond = rcond;    
    setGeneralTab();

    get_dataset(&dataset);
    get_back_settings(&backriet);

    ui->dataPowderWidget->setData(dataset, backriet);
    setPhaseTab();

    mWindow->saveEnabledActions();
    mWindow->enableActions(MainWindow::DialogOpenAction);
}

void RietveldDialog::setGeneralTab()
{
    //Automatic procedures
    ui->autoProfCheckBox->setChecked(rietCond.auto_prof);
    ui->autoStructCheckBox->setChecked(rietCond.auto_stru);
    ui->nCycleSpinBox->setValue(rietCond.ncauto);

    //Print options
    ui->startEndCheckBox->setChecked(rietCond.printOptions[StartEndCycle]);
    ui->eachCycleCheckBox->setChecked(rietCond.printOptions[EachCycle]);
    ui->correlationCheckBox->setChecked(rietCond.printOptions[CorrelMatrix]);
    ui->lsqMatrixCheckBox->setChecked(rietCond.printOptions[LSQMatrix]);
    ui->cifCheckBox->setChecked(rietCond.printOptions[CreateCIF]);

    //LSQ Options
    wSchemeChanged = false;
    ui->maxCycleSpinBox->setValue(rietCond.maxcy);
    ui->convSpinBox->setValue(rietCond.eps);
    ui->weightSchemeComboBox->setCurrentIndex(rietCond.typewei);

    //Le Bail
    ui->lebailCheckBox->setChecked(rietCond.raction);
}

void RietveldDialog::getGeneralTab()
{
    //Automatic procedures
    rietCond.auto_prof = ui->autoProfCheckBox->isChecked();
    rietCond.auto_stru = ui->autoStructCheckBox->isChecked();
    rietCond.ncauto = ui->nCycleSpinBox->value();

    //Print options
    rietCond.printOptions[StartEndCycle] = int(ui->startEndCheckBox->isChecked());
    rietCond.printOptions[EachCycle] = int(ui->eachCycleCheckBox->isChecked());
    rietCond.printOptions[CorrelMatrix] = int(ui->correlationCheckBox->isChecked());
    rietCond.printOptions[LSQMatrix] = int(ui->lsqMatrixCheckBox->isChecked());
    rietCond.printOptions[CreateCIF] = int(ui->cifCheckBox->isChecked());

    //LSQ options
    rietCond.maxcy = ui->maxCycleSpinBox->value();
    rietCond.eps = ui->convSpinBox->value();
    rietCond.typewei = ui->weightSchemeComboBox->currentIndex();

    //Le Bail?
    rietCond.raction = ui->lebailCheckBox->isChecked();

    set_refcond(&rietCond);
}

void RietveldDialog::setPhaseTab()
{
    int n = crystal_number();
    int numPhase0 = numPhase;
    if (n > numPhase) {
        for (int i=numPhase0; i < n; i++) {
            addPhaseTab();
        }
    } else if (n < numPhase) {
        for (int i = n; i < numPhase0; i++) {
            removePhaseTab();
        }
    }

    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        phaseW->setWidget(i+1);
    }

    if (numPhase > 1) {
        for (int i = 0; i < numPhase; i++) {
            ui->rietveldTabWidget->setTabText(firstPhaseTab+i, crydata::crystalPhaseName(i+1));
        }
    } else {
        ui->rietveldTabWidget->setTabText(firstPhaseTab, "Crystal");
    }
}

void RietveldDialog::getPhaseTab()
{
    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        phaseW->getData();
    }
}

void RietveldDialog::getPowderData()
{
    ui->dataPowderWidget->getData(dataset);
    set_dataset(&dataset);
}

void RietveldDialog::addPhaseTab(const QString &phaseName)
{
    auto phaseW = new PhaseWidget();
    ui->rietveldTabWidget->addTab(phaseW, phaseName);
    //ui->rietveldTabWidget->insertTab(ui->rietveldTabWidget->count(), new PhaseWidget(), phaseName);
    numPhase++;
}

void RietveldDialog::removePhaseTab()
{
    QWidget *widget = ui->rietveldTabWidget->widget(numPhase+1);
    delete widget;
    numPhase--;
}

bool RietveldDialog::hasParameterToRefine()
{    
    if (ui->autoProfCheckBox->isChecked()) return true;
    if (ui->autoStructCheckBox->isChecked()) return true;
    if (ui->dataPowderWidget->hasParameterToRefine()) return true;
    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        if (phaseW->hasParameterToRefine()) return true;
    }

    return false;
}

bool RietveldDialog::checkParameterChanged()
{
    bool okChange = false;

    poChanged.resize(numPhase);
    cellChanged.resize(numPhase);

    if (wSchemeChanged) okChange = true;
    if (ui->dataPowderWidget->hasParameterChanged()) okChange = true;
    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        if (phaseW->hasProfileFunctionChanged()) okChange = true;
        poChanged[i] = phaseW->hasPOChanged();
        cellChanged[i] = phaseW->hasCellChanged();
        qInfo() << "POChanged: " << poChanged[i] << " cellChanged: " << cellChanged[i];
        if (poChanged[i]) okChange = true;
        if (cellChanged[i]) okChange = true;
    }

    return okChange;
}

void RietveldDialog::applyChanges()
{
    bool modzero = ui->dataPowderWidget->hasParameterChanged();
    int *modcell = new int[numPhase];
    int *modfun = new int[numPhase];
    int *modpo = new int[numPhase];
    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        modcell[i] = cellChanged[i];
        modfun[i] = phaseW->hasProfileFunctionChanged();        
        modpo[i] = poChanged[i];
    }
    c_applyriet(modzero, modcell, wSchemeChanged, modfun, modpo);
}

void RietveldDialog::setLeBailRefineEnabled(bool enabled)
{
    ui->autoStructCheckBox->setEnabled(enabled);
    if (!enabled) {
        ui->autoStructCheckBox->setChecked(false);
        nCycleSpinEnabled();
    }

}

void RietveldDialog::onButtonBoxClicked(QAbstractButton *button)
{
    QDialogButtonBox::StandardButton stdButton = ui->buttonBox->standardButton(button);

    if(stdButton == QDialogButtonBox::Apply) {
        //
        //FIX - vedi with rescale
        //
        bool paramChanged = checkParameterChanged();
        bool paramToRefine = hasParameterToRefine();
        if (paramChanged || paramToRefine) {            
            getGeneralTab();
            getPhaseTab();
            getPowderData();
        }
        if (paramChanged) {
            qInfo() << "Some parameters were changed";
            applyChanges();
        }
        if (paramToRefine) {
            qInfo() << "Some parameters were checked for refinement";
            this->hide();
            QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
            finishriet_glob(&rietCond, &dataset);
            get_back_settings(&backriet);

            ui->dataPowderWidget->setData(dataset, backriet);
            setPhaseTab();

            QApplication::restoreOverrideCursor();
            this->show();
            this->activateWindow();
        } else {
            if (paramChanged)
                QMessageBox::critical(this,"Error","User modification applied. User refinement stage: please set the strategy before going on!");
            else
                QMessageBox::critical(this,"Error","User refinement stage: please set the strategy before going on!");
        }

    } else if (stdButton == QDialogButtonBox::Close) {
        c_exitriet();

    } else if (stdButton == QDialogButtonBox::Help) {
        mWindow->openHelp("rietveld-refinement");
    }
}

void RietveldDialog::autoProfileClicked(bool state)
{
    ui->dataPowderWidget->setRefineEnabled(!state);
    nCycleSpinEnabled();

    for (int i = 0; i < numPhase; i++) {
        PhaseWidget *phaseW = qobject_cast<PhaseWidget *>(ui->rietveldTabWidget->widget(firstPhaseTab+i));
        phaseW->setRefineEnabled(!state);
    }
}

void RietveldDialog::nCycleSpinEnabled()
{
    if (ui->autoProfCheckBox->isChecked() || ui->autoStructCheckBox->isChecked())
        ui->nCycleSpinBox->setEnabled(true);
    else
        ui->nCycleSpinBox->setEnabled(false);
}
